/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include <float.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
*   Here we compute the Backward part of the Forward-Backward HMM Algorithm.
*
*   Author: Adrian V. Dalca
*******************************************************************************/
VA_DOUBLE varid_backward(
    VA_MatDouble   trans_mat, 
    VA_MatDouble   e, 
    VA_VecI32      window,
    VA_MatDouble * b, 
    VA_VecDouble * s)
{
    int  i, j;
    size_t sz = reference_wgaps->size;


    //VA_ASSERT(b != NULL);
    //VA_ASSERT(s != NULL);


    //Full backward algo 
    *s = va_allocVecDouble(sz);
    for (i=0; i<sz; i++) {
	s->data[i]=-1.0/0.0;
    }
    *b = va_allocMatDouble(sz, trans_mat.size_y,"b",__FILE__,__LINE__);
    memset(b->data, 0, sizeof(VA_DOUBLE) * b->size_x * b->size_y);
    int start = window.data[0]-1; int end = window.data[1];
    //get the end in
    s->data[end -1 ] = 0;
    for (i=0; i<1600; i++) {
	//b->data[i*b->size_x+end-1]=0;
	b->data[(end-1)*1600 + i]=0;
    }
    VA_VecDouble temp = va_allocVecDouble(1600); 
    int t; int scale=(end-1-1-start)/10;
    fprintf(stderr,"   - Computing ~ ");
    for (t=end-1-1; t>=start; t--) {
        if (t%scale==0 && t/scale>1) {
                fprintf(stderr," %.2d%c", 10*(11-t/scale),'%');
        }
	for (j=0; j<1600; j++) {
		double emission_p = e.data[ (t+1)*1600 +j];
		temp.data[j]=emission_p + b->data[(t+1)*1600 + j];
	}
	#pragma omp parallel
	{
	int from_state; double largest_value=-1.0/0.0;
	#pragma omp for
	for (from_state=0; from_state<1600; from_state++) {
		//compute the f values for this position
		int to_state;
		double summands[outgoing_used[from_state]];
		double max_summand=-1.0/0.0;
		int i;
		for (i=0; i<outgoing_used[from_state]; i++) {
			to_state=outgoing[from_state][i];
			/*if (isinf(trans_mat.data[ from_state * 1600 + to_state ])) {
				fprintf(stderr,"A fatal error has occured in backward algorithm\n");
				exit(1);
			}*/
			summands[i]=trans_mat.data[ from_state * 1600 + to_state] + temp.data[to_state];
			max_summand=MAX(max_summand,summands[i]);
		}
                //optimize the log addition
		double exp_sum=0.0;
		for (i=0; i<outgoing_used[from_state]; i++) {
			double diff=max_summand-summands[i];
			if (diff<log_cutoff) {
				exp_sum+=exp(-diff);
			}
		}
		double transistion_product;
		if (exp_sum!=0.0) {
			transistion_product=log(exp_sum)+max_summand;
		} else {
			transistion_product=max_summand;
		}

		b->data[ t*1600 + from_state ] = transistion_product;
		//s->data[ t ] = MAX(s->data[t],b->data[ t*1600 + from_state ]);
		largest_value=MAX(largest_value,b->data[ t*1600 + from_state ]);
	}
	#pragma omp critical
	{
		if (s->data[ t ]<largest_value) {
			s->data[ t ]=largest_value;
		}
	}
	}
	int from_state;
	for (from_state=0; from_state<1600; from_state++) {
		b->data[ t*1600 + from_state ] -= s->data[t];
	}
    }
    free(temp.data);
    fprintf(stderr," 100%c\n",'%');
    //rc = va_GetTime(&te);
    //VA_ASSERT(rc == 0);

   // tdiff = va_TimeSub(&te, &ts);
   // fprintf(stderr, "\t\tbwd algo done in: %ld.%03ld secs\n",
   //     tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);

    return 0;



}
