/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include <float.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"


char* state_string(VA_UINT8 * states, int state) {
	char * s_state = (char*) malloc(6) ;
	//VA_ASSERT(s_state!=NULL);
	VA_VecU8 temp; temp.data=states+state*4; temp.size=4;
	temp=nuc2nuc(temp);
	sprintf(s_state,"%c%c/%c%c", temp.data[0],temp.data[1],temp.data[2],temp.data[3]);
	free(temp.data);
	return s_state;
}

/******************************************************************************
*   Here we compute the Forward part of the Forward-Backward HMM Algorithm.
*
******************************************************************************/
VA_DOUBLE varid_forward(
    VA_MatDouble   trans_mat, 
    VA_MatDouble   e, 
    VA_VecI32      window,
    VA_MatDouble * f,
    VA_VecDouble * s)
{
    int i, j;
    size_t sz = reference_wgaps->size;

    //VA_ASSERT(f != NULL);
    //VA_ASSERT(s != NULL);


    //Full forward algo
    *s = va_allocVecDouble(sz);
    for (i=0; i<sz; i++) {
	s->data[i]=-1.0/0.0;
    } 
    *f = va_allocMatDouble(sz, trans_mat.size_y,"f",__FILE__,__LINE__);
    memset(f->data, 0, sizeof(VA_DOUBLE) * f->size_x * f->size_y);
    int start = window.data[0]-1; int end = window.data[1];
    //get the begining in
    s->data[start] = 0;
    for (i=0; i<40; i++) {
	for (j=0; j<40; j++) {
		int state = i*40 + j;
		if ((i<4 || i==16) && (j<4 || j==16)) {
			//f->data[state*f->size_x+start]=e.data[start*1600+state]-log(1600);
			f->data[start*1600 + state]=e.data[start*1600+state]-log(1600);
			s->data[start]=MAX(s->data[start],f->data[start*1600 + state]);
		} else {
			//f->data[state*f->size_x+start]=-FLT_MAX;
			f->data[start*1600 + state]=-1.0/0.0;
		}
	}
    }
    for (i=0; i<1600; i++) {
	f->data[start*1600+i]-=s->data[start];
    }
    int t; int scale=(end-start-1)/10;
    fprintf(stderr,"   - Computing ~ ");
    for (t=start+1; t<end; t++) {
	if (t%scale==0 && t/scale<10) {
		fprintf(stderr," %.2d%c", 10*t/scale,'%');
	}
	#pragma omp parallel
	{
	int to_state; double largest_value=-1.0/0.0;
	#pragma omp for
	for (to_state=0; to_state<1600; to_state++) {
		//compute the f values for this position
		double summands[incoming_used[to_state]];
		double max_summand=-1.0/0.0;
		int i;
		for (i=0; i<incoming_used[to_state]; i++) {
			int from_state=incoming[to_state][i];
			/*if (isinf(trans_mat.data[ from_state * 1600 + to_state ])) {
				fprintf(stderr,"A fatal error has occured in the forward algorithm");
				exit(1);
			}*/
			//summands[i]=f->data[ from_state * f->size_x + (t-1)] + trans_mat.data[ from_state * 1600 + to_state];
			summands[i]=f->data[ (t-1)* 1600 + from_state] + trans_mat.data[ from_state * 1600 + to_state];
			max_summand=MAX(max_summand,summands[i]);
		}
		//optimize the log addition
		double exp_sum=0.0;
		for (i=0; i<incoming_used[to_state]; i++) {
			double diff=max_summand-summands[i];
			if (diff<log_cutoff) {
				exp_sum+=exp(-diff);
			}
		}
		double transistion_product;
		if (exp_sum!=0.0) {
			transistion_product=log(exp_sum)+max_summand;
		} else {
			transistion_product=max_summand;
		}
		
		double emission_p = e.data[ t*1600 +to_state ];
		/*if (isnan(emission_p)) {
			fprintf(stderr,"varid_forward : fatal error, emission probability is nan!\n");
			exit(1);
		}*/
		f->data[ t*1600 + to_state ] = emission_p + transistion_product;
		largest_value=MAX(largest_value,f->data[ t*1600 + to_state]);
		//s->data[ t ] = MAX(s->data[t],f->data[ t*1600 + to_state]);
	}
	#pragma omp critical
	{
		if (s->data[t]<largest_value) {
			s->data[t]=largest_value;
		}
	}
	}
	for (i=0; i<1600; i++) {
		f->data[ t*1600 + i ] -= s->data[t];
	}
    }
    fprintf(stderr," 100%c\n",'%');
    /* print time stats */
    //rc = va_GetTime(&te);
   // VA_ASSERT(rc == 0);

    //tdiff = va_TimeSub(&te, &ts);
    //fprintf(stderr, "\t\tfwd algo done in: %ld.%03ld secs\n",
       // tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
 

   /*for(i=0; i<20; i++) {
	printf("%3d %s\n",i,state_string(maps.full_letter.data,i));
   }*/

    return 0;


}
