/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

static void align_seq(char * name, VA_VecI32 ins, VA_VecI32 ins_vec, VA_VecU8 * seq, VA_VecU8 * qvals, VA_UINT32 * chrstart, VA_UINT32 * chrend, int color_space, char strand) { 
	    if (seq->size==0) { 
		return ;
	    }
	    //if (seq->size==1) {
	 //printf("%d\n",seq->data[0]);
	    //}
	    //Find out how much chromsome length this read actually spans
	    int chr_length=*chrend-*chrstart+1;

	    //prepare a indicator vector for gaps and inserts
	    //where not only indicate but specify length
	    VA_VecI32 gaps=va_allocVecI32(chr_length+1);
	    memset(gaps.data,0,sizeof(VA_UINT32)*gaps.size);
	    VA_VecI32 inserts=va_allocVecI32(chr_length+1);
	    memset(inserts.data,0,sizeof(VA_UINT32)*inserts.size);

	    //Find out what global inserts should be in this read
	    //along with how many global inserts happen before, (delay)
	    int j; int delay=0;
	    for (j=0; j < ins_vec.size; j++)  {
		    int pos=ins_vec.data[j];
		    //if (pos < *chrstart - front_linker_offset) {
		//delay++;
		//    } else if (pos < *chrend + back_linker_offset) {
		//	gaps.data[pos-*chrstart+1]++;
		  //  }
		    if (pos <= (*chrstart-1)) {
			delay++;
		    }
		    if (pos >= (*chrstart-1) && pos <= *chrend) {
			gaps.data[pos-(*chrstart-1)]++;
		    }
	    } 
	    //Subtract from the global inserts the local inserts
	    //populate the local inserts
	    for (j=0; j < ins.size; j++) {
		    gaps.data[ins.data[j]]--;
		    inserts.data[ins.data[j]]++;
	    }
	    delay-=inserts.data[0];
	    if (color_space==0) {
		gaps.data[0]=0; gaps.data[gaps.size-1]=0;
	    } else if (strand=='+') {
		delay-=gaps.data[0];
		gaps.data[gaps.size-1]=0;
	    } else {
		gaps.data[0]=0;
	    }
	    //count how many new bytes we need :)
	    int snp_incr=0;
	    for (j=0; j < gaps.size; j++) {
		 snp_incr+=gaps.data[j];
	    }

	    //populate the new sequence
#ifndef NDEBUG
	    VA_ASSERT(seq->size == qvals->size);
#endif
            VA_VecU8 new_seq = va_allocVecU8(seq->size+snp_incr);
            VA_VecU8 new_qvals = va_allocVecU8(qvals->size+snp_incr);
	    int newseq_index=0; int inserts_so_far=0;
 	    for (j=0; j<chr_length; j++) {
		//prefer to put inserts in first
		while (inserts.data[j]>0) {
			new_seq.data[newseq_index]=seq->data[j+inserts_so_far];
			new_qvals.data[newseq_index++]=qvals->data[j+inserts_so_far++];
			inserts.data[j]--;	
		}
		//now put in gaps
		while (gaps.data[j]>0) {
			new_seq.data[newseq_index]=4;
			new_qvals.data[newseq_index++]=GAP_QVAL;
			gaps.data[j]--;	
		}
		//now put in the next letter
		new_seq.data[newseq_index]=seq->data[j+inserts_so_far];
		new_qvals.data[newseq_index++]=qvals->data[j+inserts_so_far];
	    }
	    while (inserts.data[j]>0) {
		new_seq.data[newseq_index]=seq->data[j+inserts_so_far];
		new_qvals.data[newseq_index++]=qvals->data[j+inserts_so_far++];
		inserts.data[j]--;	
	    }
	    while (gaps.data[j]>0) {
		new_seq.data[newseq_index]=4;
		new_qvals.data[newseq_index++]=GAP_QVAL;
		gaps.data[j]--;	
	    }
	    if (j+inserts_so_far!=seq->size || newseq_index!=new_seq.size) {
		fprintf(stderr,"Read : %s, has caused an error in read re-alignment!\n",name);
		exit(1);
	    }
	    free(seq->data);
	    free(qvals->data);
	    *seq = new_seq;
	    *qvals = new_qvals;
	    //adjust the chrstart and chrend for the new reference
	    *chrstart+=delay;
	    *chrend+=delay+snp_incr+ins.size;
	    //printf("Chr start: %d\tChr end: %d, %d\n",*chrstart, *chrend,seq->size);
	    //printf("%s\n",name);
	    if (realignments_file!=NULL ) {
		    for (j=0; j<*chrstart-1; j++) {
			fprintf(realignments_file," ");
		    }
		    for (j=0; j<new_seq.size; j++) {
			fprintf(realignments_file,"%d",new_seq.data[j]);
		    }
		    fprintf(realignments_file,"\t\t\t%s\n",name);
	    	    for (j=0; j<*chrstart-1; j++) {
	    	    	fprintf(realignments_file," ");
	    	    }
	    }
	    if (gaps.size>0) {
		free(gaps.data);
	    }
	    if (inserts.size>0) {
		free(inserts.data);
	    }
	    return;
}

/******************************************************************************
* aligns the mapped reads and the reference
*   [new_ref, reads] = ALIGN_READS(ref, reads, ins_vec) aligns the mapped
*   reads and the reference in the very rare case of a gap present in the
*   mapped read and not present in the reference or another read at that
*   position.
*
*   Parameters:
*   ref - the reference in ns letterspace
*   reads - the read vector struct
*   ins_vector - a vector indicating places in the reference where an
*       insertion occurs due to a mapping
*
*   Returns:
*   new_ref - the aligned reference
*   reads - the aligned reads
*
*   Author: Adrian V Dalca
******************************************************************************/
VA_Vector align_reads(
    VA_Vector reads,
    VA_VecI32 ins_vec)
{    
    int i;
    ReadsNode_t * rd = (ReadsNode_t *)reads.data;
    
    /******************
     * New method
     ******************/


    fprintf(stderr, "   * There are %lu potential insert sites.\n",ins_vec.size);
    if (realignments_file != NULL ){
	fwrite(reference_wgaps->sequence,1,reference_wgaps->size,realignments_file);
	fprintf(realignments_file,"\n ");
	for (i=0; i<reference_wgaps->size; i++) {
		fprintf(realignments_file,"%u",reference_wgaps->colour_sequence[i]);
	}
	fprintf(realignments_file,"\n");
    }
        for (i = 0; i < reads.size; i++) { 
            /* get any insertions that are between [s, fi)
             * these bounds are chosen specifically so because we care about
             * an insertion @ s (since it would be put after s) but not about
             * one @ fi (since it would be put AFTER fi, and hence not in our
             * care
             */
	    //printf("Read %s\n",rd[i].name);
	    //printf("LSseq start %d end %d inserts size %d\n", rd[i].ls_chrstart, rd[i].ls_chrend, rd[i].ls_ins.size);
	    //printf("CSseq start %d end %d inserts size %d\n", rd[i].cs_chrstart, rd[i].cs_chrend, rd[i].cs_ins.size);
	    int j;
	    align_seq(rd[i].name, rd[i].ls_ins, ins_vec, &rd[i].ls_seq, &rd[i].ls_qvals, &rd[i].ls_chrstart, &rd[i].ls_chrend, 0 , rd[i].strand);
	    if (realignments_file!=NULL && rd[i].ls_seq.size!=0) {
		    for (j=0; j<rd[i].ls_qvals.size; j++) {
			fprintf(realignments_file,"%c",rd[i].ls_qvals.data[j]);
		    }
	    	    fprintf(realignments_file,"\n");
	    }
	    //printf("Colseq\n");
	    align_seq(rd[i].name, rd[i].cs_ins, ins_vec, &rd[i].cs_seq, &rd[i].cs_qvals, &rd[i].cs_chrstart, &rd[i].cs_chrend, 1 , rd[i].strand); 
	    if (realignments_file!=NULL && rd[i].cs_seq.size!=0) {
		    for (j=0; j<rd[i].cs_qvals.size; j++) {
			fprintf(realignments_file,"%c",rd[i].cs_qvals.data[j]);
		    }
		    fprintf(realignments_file,"\n");
	    }
	    
        }

        //rc = va_GetTime(&te);
        //VA_ASSERT(rc == 0);

        //tdiff = va_TimeSub(&te, &areadtic);

        /* done read prep */
        //fprintf(stderr, "\t Done in: %ld.%03ld secs \n",
          //  tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
    //}
    //exit(1);
    return (reads);
}


