/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include "va_assert.h"
#include "va_log.h"
#include "varid.h"
#include "sam2pretty_lib.h"
#include <errno.h>


long long unsigned split_reads() {


	FILE * remapped_file = fopen(remapped_filename,"w");
	if (remapped_file==NULL) {
		fprintf(stderr,"There has been an error in opening the file %s for writting\n",remapped_filename);
		perror("");
		exit(1);
	}

	if (predict_start!=0) {
		pretty_remap_header(remapped_file,reference->name,hmm_start,hmm_end);
	} else {
		pretty_header(remapped_file,reference->name,reference->size);
	}

	long long unsigned totreads = 0;
	long long unsigned usedreads = 0;

	/* go through each line and re-write it in the right file */
	/* get the line (with newline char) */
	char line[BUFSIZ];
	while (fgets(line, BUFSIZ, alignments_file) != NULL) {
		if (line[0] == '@') {
			continue;
		}

		//Found a read
		totreads++;
		
		//find the contig name, skip first two tabs
		if (predict_start==0) {
			int i; int length=strlen(line);
			for(i=0; i<length && line[i]!='\t'; i++); i++;
			for(; i<length && line[i]!='\t'; i++); i++;
			if (i>=length) {
				fprintf(stderr,"A fatal error has occured in parsing SAM input!\n");
				exit(1);
			}
			char * rname = line+i;
			int min_length=MIN(strlen(reference->name),strlen(rname));
			if (strncmp(rname,reference->name,min_length)==0) {
				pretty * pa = pretty_from_string(line);
				if (pa->mapped) {
					pretty * head = pretty_sub_prettys(pa);
					while ( head != NULL ) {
						//pretty_ls(head);
						//pretty_qualities(head);
						//if (pa->colour_space) {
						//	pretty_cs(head);
						//}
						pretty_print_sam(remapped_file,head);
						pretty * temp=(pretty*)head->next;
						if (head==pa) {
							pa=NULL;
						}
						pretty_free(head);
						head=temp;
						usedreads++;
					}
					if (pa!=NULL) {
						pretty_free(pa);
					}
				} else {
					pretty_free(pa);
				}
			}
		} else {
			pretty * pa=pretty_from_string(line);	
			if (pa!=NULL) {
				if (pa->mapped) {
					int res = pretty_remap(pa,hmm_start,hmm_end);
					if (res==1) {
						/* determine where this read belongs */
							int min_length=MIN(strlen(reference->name),strlen(pa->reference_name));
							if (strncmp(pa->reference_name,reference->name,min_length)==0) {
								pretty * head = pretty_sub_prettys(pa);
								while ( head != NULL ) {
									//pretty_ls(head);
									//pretty_qualities(head);
									//if (pa->colour_space) {
									//	pretty_cs(head);
									//}
									pretty_print_sam(remapped_file,head);
									if (mix_signals && head->colour_space) {
										pretty_print_sam_force_ls(remapped_file,head);
									}
									pretty * temp=(pretty*)head->next;
									if (head==pa) {
										pa=NULL;
									}
									pretty_free(head);	
									head=temp;
									usedreads++;
								}
								if (pa!=NULL) {
									pretty_free(pa);
								}
							}
					} else {
						pretty_free(pa);
					}
				} else {
					pretty_free(pa);
				}
			}
		}
	}

	fclose(remapped_file);
	/* oprint some useful information */
	fprintf(stderr, "   * Total reads: %llu\n", totreads);
	fprintf(stderr, "   * Used reads : %llu\n", usedreads);

	return usedreads;
}
